/*
 *   Copyright (C) 2011-2012 Bixolon Co., Ltd.
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *        
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *             
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * BIXOLON CO LTD
 *
 * Thanks to people who are involved in this project.

 * Developed By dwlee  May-24-2012 Bixolon Co Ltd, South Korea
 * Mail: dwlee@bixolon.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <cups/http.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/lp.h>
#include "bxlMsq.h"
#include "bxlApiMsqKey.h"

#include <bluetooth/bluetooth.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#define IOCNR_GET_DEVICE_ID 1
#define LPIOC_GET_DEVICE_ID(len)    _IOC(_IOC_READ, 'P', IOCNR_GET_DEVICE_ID, len)

#define MAXMSQLEN 4096

enum {					/**** Backend exit codes ****/
	CUPS_BACKEND_OK = 0,		/* Job completed successfully */
	CUPS_BACKEND_FAILED = 1,	/* Job failed, use error-policy */
	CUPS_BACKEND_AUTH_REQUIRED = 2,	/* Job failed, authentication required */
	CUPS_BACKEND_HOLD = 3,		/* Job failed, hold job */
	CUPS_BACKEND_STOP = 4,		/* Job failed, stop queue */
	CUPS_BACKEND_CANCEL = 5,	/* Job failed, cancel job */
	CUPS_BACKEND_RETRY = 6,		/* Failure requires us to retry (BlueZ specific) */
};

static int readMsqID 	= -1;
static int writeMsqID	= -1;
static int debug_mode 	= 0;

// XXX_FILTER_AGENT_MSQKEY
static int readKey  = FILTER_AGENT_MSQKEY;
// XXX_AGENT_API_MSQKEY
static int writeKey = AGENT_API_MSQKEY; 

int USBOpenDevice (const char *uri);
int SerialOpenDevice (const char *uri);
int ParallelOpenDevice (const char *uri);
int SocketOpenDevice (const char *uri);
int BluetoothOpenDevice (const char *uri);

int getUSBDeviceURI (int fd, char *device_uri, int uriLen);
int getDeviceOption (
		const char *device_id, 
		const char *option, 
		char *buffer);

int compURI(const char *uri, const char *device_uri);

int spp_connect (bdaddr_t *src, bdaddr_t *dst, uint8_t channel, 
				const char *cups_class, int *sk);
int sdp_search_spp(sdp_session_t *sdp, uint8_t *channel);

int main (int argc, char *argv[])
{
	int ASBMode;
	int StatMode;
	int msgflag;
	int nread;
	int temp_fd;
	int device_fd;
	int asb_fd;
	int len;
	int interface;
	struct bxlMsg_t *msg;

	char uri[1024];
	char *uri_p;
	char cmd[10];
	char *cmd_p;

	fd_set input;
	struct timeval timeout;
	int ret;
	int offline;
	char answer[5];

	// 보안 메시지. CUPS Driver 쪽에서 uri, cmd를 전송전에 이 msg를 전송하지 
	// 않으면 fail 처리
	char BxlAPISecurityMsg[50] = "@THI\0$l$81\x10x@l0\0n$\x1D\x77PI\0M59\x10\x04";

	if (argc == 2) {
		debug_mode = atoi(argv[1]);
	}

	ASBMode 	= 0;
	StatMode	= 0;
	device_fd 	= -1;
//	asb_fd		= -1;
	offline		= 0;
	len			= 0;
	interface   = 0;

	memset(uri, 0, sizeof(uri));
	memset(cmd, 0, sizeof(cmd));

	if ((readMsqID = bxlMsqOpen(readKey, MAXMSQLEN)) == EXIT_FAILURE) {
		printf("bxlStatusChkAgent : read bxlMsqOpen() fail.\n");

		return EXIT_FAILURE;
	}   

	if ((writeMsqID = bxlMsqOpen(writeKey, MAXMSQLEN)) == EXIT_FAILURE) {
		printf("bxlStatusChkAgent : write bxlMsqOpen() fail.\n");

		return EXIT_FAILURE;
	} 

	if (debug_mode) {
		printf("DEBUG : MSQ Open Success.\n");
		printf("DEBUG : readMsqID = %d, writeMsqID = %d\n", readMsqID, writeMsqID);
	}

	if ((msg = calloc(1, sizeof(long) + MAXMSQLEN + 1)) == NULL) {
		printf("bxlStatusChkAgent : fail memory allocation.\n");

		return EXIT_FAILURE;
	}

	while (1) {
		if (debug_mode) {
			printf("ASBMode = %d\n", ASBMode);
		}

		if (ASBMode == 1) {
			// ASB가 set 되어 있으면 바로 바로 리턴하도록
			// pending되면 printer로 부터 read를 할 수 없으므로
			msgflag = IPC_NOWAIT;
		} else {
			msgflag = 0;
		}

		// security message
		memset(msg, 0, sizeof(msg));
		if ((nread = msgrcv(readMsqID, msg, MAXMSQLEN, readKey, msgflag)) < 0) {
			if (ASBMode != 1) {
				printf("bxlStatusChkAgent : fail receive message.\n");
			}

			//return EXIT_FAILURE;

			if (ASBMode == 1) {
//				device_fd = asb_fd;
				goto receive_status;
			} else {
				continue;
			}
		} else {
			if (memcmp(msg->mtext, BxlAPISecurityMsg, 28)) {
				printf("bxlStatusChkAgent : not support bixolon printer.\n");

				return EXIT_FAILURE;
			}
		}

		// receive uri
		memset(msg, 0, sizeof(msg));
		if ((nread = msgrcv(readMsqID, msg, MAXMSQLEN, readKey, msgflag)) < 0) {
//			printf("bxlStatusChkAgent : fail receive message.\n");

			//return EXIT_FAILURE;
			if (ASBMode == 1) {
//				device_fd = asb_fd;
				goto receive_status;
			} else {
				continue;
			}
		} else {
			msg->mtext[nread] = '\0';

			memset(uri, 0, sizeof(uri));
			strcpy(uri, msg->mtext);
			uri_p = uri + strspn(uri, " \t");
		}

		if (debug_mode) {
			printf("DEBUG : receive uri = %s\n", uri_p);
		}

		memset(msg, 0, sizeof(msg));
		if ((nread = msgrcv(readMsqID, msg, MAXMSQLEN, readKey, msgflag)) < 0) {
//			printf("bxlStatusChkAgent : fail receive message.\n");

			//return EXIT_FAILURE;
			if (ASBMode == 1) {
//				device_fd = asb_fd;
				goto receive_status;
			} else {
				continue;
			}
		} else {
			if (memcmp(msg->mtext, BxlAPISecurityMsg, 28)) {
				printf("bxlStatusChkAgent : not support bixolon printer.\n");

				return EXIT_FAILURE;
			}
		}

		// receive cmd
		memset(msg, 0, sizeof(msg));
		if ((nread = msgrcv(readMsqID, msg, MAXMSQLEN, readKey, msgflag)) < 0) {
//			printf("bxlStatusChkAgent : fail receive message.\n");

			//return EXIT_FAILURE;
			if (ASBMode == 1) {
//				device_fd = asb_fd;
				goto receive_status;
			} else {
				continue;
			}
		} else {
			msg->mtext[nread] = '\0';

			memset(cmd, 0, sizeof(cmd));
			strcpy(cmd, msg->mtext);
			cmd_p = cmd + strspn(cmd, " \t");

			if (debug_mode) {
				printf("\nDEBUG : receive cmd = 0x%x 0x%x 0x%x\n", cmd_p[0], cmd_p[1], cmd_p[2]);
			}

			// 중복되는 GSa명령 처리
			if (!memcmp(cmd_p, "\x1D\x61", 2)) { 

				if (ASBMode == 1 && (cmd_p[2] >= 0 && cmd_p[2] < 256)) {
//					device_fd = asb_fd;

					goto receive_status;
				} else if (ASBMode == 0 && cmd_p[2] == 0) {
					// ASB가 설정되어 있지 않은 상태에서 asb mode off 명령이 오면 무시
					continue;
				}
			}
		}

		// check cmd
		if ((memcmp(cmd_p, "\x10\x04", 2) 		&&
			 memcmp(cmd_p, "\x04", 1)		 	&&
			 memcmp(cmd_p, "\x1D\x49", 2) 	 	&&
			 memcmp(cmd_p, "\x1D\x49\x62", 3) 	&&
			 memcmp(cmd_p, "\x1D\x61", 2) 	 	&&
			 memcmp(cmd_p, "\x1D\x72", 2)) 	 	||
			 strlen(cmd_p) < 2 || strlen(cmd_p) > 3) {
			printf("bxlStatusChkAgent : Doesn't support command.\n");

			if (ASBMode != 1) {
				continue;
			}
		}

receive_status:

		//+++++++++++++++++++++++ 
		bxlMsqClean(writeMsqID);

		// LSBMode 이든 아니든 device_fd는 매 루프마다 열고 닫는다.
		
//		if (ASBMode == 0) {
			// printer의 uri를 이용하여 printer의 
			if (!strncasecmp(uri_p, "usb:", 4)) {
				if ((device_fd = USBOpenDevice(uri_p)) < 0) {
					printf("bxlStatusChkAgent : failed to open USB printer.\n");

					continue;
				}
			} else if (!strncasecmp(uri_p, "serial:", 7)) {
				if ((device_fd = SerialOpenDevice(uri_p)) < 0) {
					printf("bxlStatusChkAgent : failed to open Serial printer.\n");

					continue;
				}
			} else if (!strncasecmp(uri_p, "parallel:", 9)) {
				if ((device_fd = ParallelOpenDevice(uri_p)) < 0) {
					printf("bxlStatusChkAgent : failed to open Parallel printer.\n");

					continue;
				}
			} else if (!strncasecmp(uri_p, "socket:", 7)) {
				if ((device_fd = SocketOpenDevice(uri_p)) < 0) {
					printf("bxlStatusChkAgent : failed to open socket printer.\n");

					continue;
				}
			} else if (!strncasecmp(uri_p, "bluetooth:", 10)) {
				// sleep을 안주고 빠르게 open, close를 반복하면 
				// connection fail 발생
				usleep(5000);

				//++++++++++++++++++
				// When LSB Mode cancel, for avoid bluetooth interface clash.
				if (ASBMode == 1 && cmd_p != NULL && !memcmp(cmd_p, "\x1D\x61\x00", 3)) {
					sleep(6);
				}

				if ((device_fd = BluetoothOpenDevice(uri_p)) < 0) {
					printf("bxlStatusChkAgent : failed to open bluetooth printer.\n");

					continue;
				}
			} else {
				printf("Invalid printer uri.\n");

				continue;
			}

			if (debug_mode) {
				printf("device_fd = %d\n", device_fd);
			}

//		} 
	/*	
		else {
			// ASBMode가 1인 경우에는 STATMode에서도 기존에 열려있는 
			// device_fd를 사용해야 한다. 중복해서 open 불가 
			device_fd = asb_fd;
		}
	*/

		
		if (cmd_p != NULL) {
			if ((len = write(device_fd, cmd_p, strlen(cmd_p))) < 0) {
				printf("bxlStatusChkAgent : failed to send command.\n");

				if (device_fd > 0) {
					close(device_fd);
					device_fd = -1;
				}
				continue;
			}
		}

		if (debug_mode) {
			if (cmd_p != NULL) {
				printf("DEBUG: write len = %d, cmd len = %d\n", len, strlen(cmd_p));
			}
		}

//receive_status:

		if (cmd_p == NULL) {
			// NULL인 경우 아무 비교 하지 않음
			
		} else if (!memcmp(cmd_p, "\x1D\x61", 2) && ( cmd_p[2] > 0 && cmd_p[2] < 256)) {
			printf("Enable 'Last Status Back(LSB)' function.\n");
			ASBMode = 1;
		//	asb_fd = device_fd;
		} else if (!memcmp(cmd_p, "\x1D\x61\x00", 3)) {

			if ((len = write(device_fd, cmd_p, strlen(cmd_p))) < 0) {
				printf("bxlStatusChkAgent : Failed to disable LSB.\n");

				continue;
			}

			ASBMode = 0;

			answer[0] = 0x00;
			bxlMsqSend(writeMsqID, writeKey, answer, 1);

			if (device_fd > 0) {
				close(device_fd);
				device_fd = -1;
		//		asb_fd = -1;
			}

			printf("Disable 'Last Status Back(LSB)' function.\n");

			continue;
		} else if (cmd_p != NULL) {
			StatMode = 1;
		}

		if (debug_mode) {
			if (cmd_p != NULL) {
				printf("DEBUG : receive cmd = 0x%x 0x%x 0x%x\n", cmd_p[0], cmd_p[1], cmd_p[2]);
			}
			printf("asb_fd = %d, device_fd = %d\n", asb_fd, device_fd);
			printf("DEBUG : ASBMode = %d, StatMode = %d\n", ASBMode, StatMode);
		}

		while (2) {
			// recieve status response from printer
			FD_ZERO(&input);
			FD_SET(device_fd, &input);

			if (StatMode == 0) {
				// 0초로 설정하니까 usb는 정상 동작하는데,
				// bluetooth인 경우 응답을 받지 못함.
				// command메뉴얼을 참고하여 1 sec으로 설정
				timeout.tv_sec  = 1;
				timeout.tv_usec = 0;
			} else {
				timeout.tv_sec  = 2;
				timeout.tv_usec = 5;
			}


			if (select(device_fd + 1, &input, NULL, NULL, &timeout) <= 0) {
				if (errno == ENXIO && offline < 5) {
					printf("bxlStatusChkAgent : printer is currently offline.\n");
					sleep(1);
					offline++;

					continue;
				} else if (errno == EINTR) {
					printf("bxlStatusChkAgent : interrupted this process.\n");

					goto fail_job;
				} else if (ASBMode == 1 && StatMode == 0) {
					// NO answer (ASB Mode)
					printf("ASBMode no answer.\n");
					break;
				} else if (offline < 100) {
					usleep(50000);
					offline++;
					printf("offline.\n");

					continue;
				}

				printf("bxlStatusChkAgent : failed to receive answer.\n");

				if (debug_mode) {
					printf("bxlStatusChkAgent : %s\n", strerror(errno));
				}

				if (device_fd > 0) {
					close(device_fd);
					device_fd = -1;
				}

				break;
			}


			if (FD_ISSET(device_fd, &input)) {
				memset(answer, 0, sizeof(answer));
				
				// recieve response
				if ((ret = read(device_fd, answer, sizeof(answer))) > 0) {

					printf("successed read status info.\n");

					if (ret == 1) {
						// GS r, DLE EOT, GS I answer
						StatMode = 0;
					} else if (ret == 4 && answer[0] == 0x5F) {
						// GS I answer
						StatMode = 0;
					} else {
						// GS a answer
					}

//					bxlMsqClean(writeMsqID);

					// send to BxlAPI read function
					if (bxlMsqSend(writeMsqID, writeKey, answer, ret) == EXIT_FAILURE) {
						printf("bxlStatusChkAgent : failed to send answer.\n");

						if (device_fd > 0) {
							close(device_fd);
							device_fd = -1;
						}
						break;
					}

					break;

				} else if (ret == 0) {
					// USB 2.0 null packet 처리
					usleep(10000);

					continue;
				}
			}

		} // END while(2)

		if (debug_mode) {
			printf("end read_send loop\n");
		}

//		if (device_fd > 0 && ASBMode == 0) {
		if (device_fd > 0 ) {
			close(device_fd);
			device_fd = -1;
		}
		
		cmd_p = NULL;
		usleep(50000);
	} // END while(1)

	return EXIT_SUCCESS;

fail_job:
	if (device_fd > 0) {
		close(device_fd);
	}

	if (msg != NULL) {
		free(msg);
	}

	return EXIT_FAILURE;
}

// SPP profile만 지원. 
int BluetoothOpenDevice (const char *uri)
{
	sdp_session_t *sdp;
	bdaddr_t bdaddr;
	const char *ptr, *cups_class;
	uint8_t b[6], channel;
	char str[3], device[18];
	int i, fd, err;
	unsigned short ctrl_psm, data_psm;

	ptr = NULL;
	memset(b, 0, sizeof(b));
	memset(device, 0, sizeof(device));
	memset(str, 0, sizeof(str));

	if (debug_mode) {
		printf("BluetoothOpenDevice() start\n");
		printf("bluetooth uri = %s\n", uri);
	}

	ptr = uri + 12;
	for (i = 0; i < 6; i++) {
		strncpy(str, ptr, 2);
		b[i] = (uint8_t)strtol(str, NULL, 16);
		ptr += 2;
	}

	sprintf(device, "%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X",
					b[0], b[1], b[2], b[3], b[4], b[5]);

	if (debug_mode) {
		printf("bluetooth printer addr : %s\n", device);
	}

	str2ba(device, &bdaddr);

service_search :
	if (!(sdp = sdp_connect(BDADDR_ANY, &bdaddr, SDP_RETRY_IF_BUSY))) {
		printf("Can't open Bluetooth connection\n");

		return -1;
	}

	err = sdp_search_spp(sdp, &channel);

	sdp_close(sdp);

	cups_class = getenv("CLASS");

	if (err) {
		if (cups_class) {
			printf("Unable to contact printer, queuing on next "
					"printer in class...\n");

//			sleep(3);

			return -1;
		}

		sleep(3);
		printf("Can't get service information\n");

		goto service_search;
	}

connect :
	err = spp_connect(BDADDR_ANY, &bdaddr, channel, 
			cups_class, &fd);

	if (err == CUPS_BACKEND_FAILED && cups_class) {
		printf("INFO: Unable to contact printer, queuing on "
				"next printer in class...\n");
//		sleep(3);

		return -1;

	} else if (err == CUPS_BACKEND_RETRY) {
		printf("Backend Retry\n");
		sleep(3);

		goto connect;
	}

	return fd;
}

int USBOpenDevice (const char *uri)
{
	int fd;
	int i;
	char device[255];
	char device_uri[1024];

	if (debug_mode) {
		printf("USBOpenDevice() : start.\n");
	}

	memset(device_uri, 0, sizeof(device_uri));

	if (!strncmp(uri, "usb:/dev/", 9)) {
		errno = ENODEV;

		return -1;

	} else if (!strncmp(uri, "usb://", 6)) {

		while (1) {
			for (i = 0; i < 16; i++) {
				sprintf(device, "/dev/usblp%d", i);

				if ((fd = open(device, O_RDWR | O_EXCL | O_NONBLOCK/*| S_IRWXG | S_IRWXU | S_IRWXO*/)) < 0 && 
						errno == ENOENT) {
					sprintf(device, "/dev/usb/lp%d", i);

					if ((fd = open(device, O_RDWR | O_EXCL| O_NONBLOCK /*| S_IRWXG | S_IRWXU | S_IRWXO*/)) < 0 &&
							errno == ENOENT) {
						sprintf(device, "/dev/usb/usblp%d", i);

						if ((fd = open(device, O_RDWR | O_EXCL| O_NONBLOCK/* | S_IRWXG | S_IRWXU | S_IRWXO*/)) < 0 &&
								errno == ENOENT) {
							continue;
						}
					}
				} // END if

				if (debug_mode) {
					printf("USBOpenDevice() : device = %s\n", device);
				}

				if (fd >= 0) {
					getUSBDeviceURI(fd, device_uri, sizeof(device_uri));
				} else {
					device_uri[0] = '\0';
				}

				if (!strcmp(uri, "\0") || !strcmp(device_uri, "\0")) 
					continue;

				if (compURI(uri, device_uri) == EXIT_SUCCESS) {
					return fd;
				} else {
					if (fd >= 0) {
						close(fd);
					}
				}
			} // END for

			sleep(1);
		} // END while(1)
	}

	errno = ENODEV;

	return -1;
}

int compURI(const char *uri, const char *device_uri)
{
	char *uri_tmp;
	char *duri_tmp;
	char uri_mnf[20];
	char uri_mdl[20];
	char duri_mnf[20];
	char duri_mdl[20];
	char *tmp;

	if (uri == NULL || device_uri == NULL) { 
		printf("compURI() : Incorrect argument.");

		return EXIT_FAILURE;
	}

	if ((duri_tmp = calloc(1, strlen(device_uri) + 1)) == NULL) {
		printf("compURI() : failed memory allocation.\n");

		return EXIT_FAILURE;
	}

	memcpy(duri_tmp, device_uri, strlen(device_uri));

	if ((uri_tmp = calloc(1, strlen(uri) + 1)) == NULL) {
		printf("compURI() : failed memory allocation.\n");

		return EXIT_FAILURE;
	}

	memcpy(uri_tmp, uri, strlen(uri));

	memset(uri_mnf, 0, sizeof(uri_mnf));
	memset(uri_mdl, 0, sizeof(uri_mdl));
	memset(duri_mnf, 0, sizeof(duri_mnf));
	memset(duri_mdl, 0, sizeof(duri_mdl));

	tmp = strstr(duri_tmp, "://");
	tmp += 3;
	
	strcpy(duri_mnf, strtok(tmp, "//"));
	strcpy(duri_mdl, strtok(NULL, "//"));
	
	tmp = strstr(uri_tmp, "://");
	tmp += 3;

	strcpy(uri_mnf, strtok(tmp, "//"));
	strcpy(uri_mdl, strtok(NULL, "//"));

	if (strncmp(duri_mnf, uri_mnf, strlen(duri_mnf)) ||
		strncmp(duri_mdl, uri_mdl, strlen(duri_mdl))) {
		free(uri_tmp);
		free(duri_tmp);

		return EXIT_FAILURE;
	}

	free(uri_tmp);
	free(duri_tmp);

	return EXIT_SUCCESS;
}

int SerialOpenDevice (const char *uri)
{
	char *options;
	char *opt_name;
	char *opt_val;
	char newURI[256];
	char *tmp;
	char sep;
	int fd;

	struct termios opts;
	struct termios origopts;

	if ((tmp = strchr(uri, ':')) == NULL) {
		printf("SerialOpenDevice() : invalid argument.\n");

		return -1;
	} else {
		tmp++;

		memset(newURI, 0, sizeof(newURI));
		strcpy(newURI, tmp);
	}


	if ((options = strchr(newURI, '?')) != NULL) {
		//*options++ = '\0';
		*options = '\0';
		options++;
	}

	if (debug_mode) {
		printf("DEBUG : serial port = *%s*\n", newURI);
	}

	if ((fd = open(newURI, O_RDWR | O_NOCTTY | O_EXCL | O_NDELAY)) == -1) {
		printf("SerialOpenDevice() : %s serial port open error.\n");

		return -1;
	}

	tcgetattr(fd, &origopts);
	tcgetattr(fd, &opts);

	opts.c_lflag &= ~(ICANON | ECHO | ISIG);
	/* Raw mode */
	opts.c_oflag &= ~OPOST;		/* Don't post-process */


	if (options) {
		while (*options) {
			/*
			 * Get the name...
			 */

			opt_name = options;

			while (*options && 
					*options != '=' && 
					*options != '+' && 
					*options != '&') { 
				options++;
			}

			if ((sep = *options) != '\0') {
				//*options++ = '\0';
				*options = '\0';
				options++;
			}

			if (sep == '=') {
				/*
				 * Get the value...
				 */

				opt_val = options;

				while (*options && *options != '+' && *options != '&') {
					options++;
				}

				if (*options) {
					//*options++ = '\0';
					*options = '\0';
					options++;
				}
			} else {
				opt_val = (char *)"";
			}

			if (debug_mode) {
				printf("DEBUG : option name = %s, value = %s\n", opt_name, opt_val);
			}

			/*
			 * Process the option...
			 */

			if (!strcasecmp(opt_name, "baud")) {
				/*
				 * Set the baud rate...
				 */

				//print_size = atoi(opt_val) / 100;

#if B19200 == 19200
				cfsetispeed(&opts, atoi(opt_val));
				cfsetospeed(&opts, atoi(opt_val));
#else
				switch (atoi(opt_val))
				{
					case 1200 :
						cfsetispeed(&opts, B1200);
						cfsetospeed(&opts, B1200);
						break;
					case 2400 :
						cfsetispeed(&opts, B2400);
						cfsetospeed(&opts, B2400);
						break;
					case 4800 :
						cfsetispeed(&opts, B4800);
						cfsetospeed(&opts, B4800);
						break;
					case 9600 :
						cfsetispeed(&opts, B9600);
						cfsetospeed(&opts, B9600);
						break;
					case 19200 :
						cfsetispeed(&opts, B19200);
						cfsetospeed(&opts, B19200);
						break;
					case 38400 :
						cfsetispeed(&opts, B38400);
						cfsetospeed(&opts, B38400);
						break;
#ifdef B57600
					case 57600 :
						cfsetispeed(&opts, B57600);
						cfsetospeed(&opts, B57600);
						break;
#endif /* B57600 */
#ifdef B115200
					case 115200 :
						cfsetispeed(&opts, B115200);
						cfsetospeed(&opts, B115200);
						break;
#endif /* B115200 */
#ifdef B230400
					case 230400 :
						cfsetispeed(&opts, B230400);
						cfsetospeed(&opts, B230400);
						break;
#endif /* B230400 */
					default :
						printf("SerialOpenDevice(): Unsupported baud rate %s!\n",
								opt_val);
						break;
				}
#endif /* B19200 == 19200 */
			}
			else if (!strcasecmp(opt_name, "bits")) {
				/*
				 * Set number of data bits...
				 */

				switch (atoi(opt_val)) {
					case 7 :
						opts.c_cflag &= ~CSIZE;
						opts.c_cflag |= CS7;
						opts.c_cflag |= PARENB;
						opts.c_cflag &= ~PARODD;
						break;
					case 8 :
						opts.c_cflag &= ~CSIZE;
						opts.c_cflag |= CS8;
						opts.c_cflag &= ~PARENB;
						break;
				}
			}
			else if (!strcasecmp(opt_name, "parity")) {
				/*
				 * Set parity checking...
				 */

				if (!strcasecmp(opt_val, "even")) {
					opts.c_cflag |= PARENB;
					opts.c_cflag &= ~PARODD;
				} else if (!strcasecmp(opt_val, "odd")) {
					opts.c_cflag |= PARENB;
					opts.c_cflag |= PARODD;
				} else if (!strcasecmp(opt_val, "none")) {
					opts.c_cflag &= ~PARENB;
				} else if (!strcasecmp(opt_val, "space")) {
					/*
					 * Note: we only support space parity with 7 bits per character...
					 */
					opts.c_cflag &= ~CSIZE;
					opts.c_cflag |= CS8;
					opts.c_cflag &= ~PARENB;
				} else if (!strcasecmp(opt_val, "mark")) {
					/*
					 * Note: we only support mark parity with 7 bits per character
					 * and 1 stop bit...
					 */

					opts.c_cflag &= ~CSIZE;
					opts.c_cflag |= CS7;
					opts.c_cflag &= ~PARENB;
					opts.c_cflag |= CSTOPB;
				}
			}
			else if (!strcasecmp(opt_name, "flow")) {
				/*
				 * Set flow control...
				 */

				if (!strcasecmp(opt_val, "none")) {
					opts.c_iflag &= ~(IXON | IXOFF);
					opts.c_cflag &= ~CRTSCTS;

				} else if (!strcasecmp(opt_val, "soft")) {
					opts.c_iflag |= IXON | IXOFF;
					opts.c_cflag &= ~CRTSCTS;

				} else if (!strcasecmp(opt_val, "hard") ||
							!strcasecmp(opt_val, "rtscts")) {
					opts.c_iflag &= ~(IXON | IXOFF);
					opts.c_cflag |= CRTSCTS;

				} else if (!strcasecmp(opt_val, "dtrdsr")) {
					opts.c_iflag &= ~(IXON | IXOFF);
					opts.c_cflag &= ~CRTSCTS;
				}
			}
			else if (!strcasecmp(opt_name, "stop")) {
				switch (atoi(opt_val)) {
					case 1 :
						opts.c_cflag &= ~CSTOPB;
						break;

					case 2 :
						opts.c_cflag |= CSTOPB;
						break;
				}
			}
		} // END while (*options)
	}

	tcsetattr(fd, TCSANOW, &opts);
	fcntl(fd, F_SETFL, 0);

	/*
	tcsetattr(fd, TCSADRAIN, &origopts);
	close(fd);
	*/

	return fd;
}

int ParallelOpenDevice (const char *uri)
{
	char *newURI;
	char *tmp;

	int fd;

	if ((newURI = strchr(uri, ':')) == NULL) {
		printf("SerialOpenDevice() : invalid argument.\n");

		return -1;
	} else {
		newURI++;
	}

	if ((fd = open(newURI, O_RDWR | O_EXCL)) < 0) {
		printf("SerialOpenDevice() : This system parallel drives currently are not support bidirection I/O.\n");

		return -1;
	}

	return fd;
}

int SocketOpenDevice (const char *uri)
{
	char hostname[1024];
	char portname[255];

	char newURI[1024];
	char *tmp;

	int fd;
	int delay;
	int error;
	int contimeout;
	int recoverable;
	time_t start_time;

	http_addrlist_t *addrlist;
	http_addrlist_t *addr;

	fd = -1;
	recoverable = 0;
 	contimeout  = 7 * 24 * 60 * 60;

	if ((tmp = strchr(uri, ':')) == NULL) {
		printf("SocketOpenDevice() : invalid argument.\n");

		return -1;
	} else {
		tmp = tmp + 3;
		memset(newURI, 0, sizeof(newURI));
		strcpy(newURI, tmp);

		if ((tmp = strchr(newURI, ':')) == NULL) {
			strcpy(hostname, newURI);
			strcpy(portname, "socket");
		} else {
			memcpy(tmp, "\0", 1);
			tmp++;

			strcpy(hostname, newURI);
			strcpy(portname, tmp);
		}
	}

	start_time = time(NULL);

	if ((addrlist = httpAddrGetList(hostname, AF_UNSPEC, portname)) == NULL) {
		printf("SocketOpenDevice() : Unable to locate printer \'%s\'.\n", hostname);

		return -1;
	}

	for (delay = 5;;) {
		if ((addr = httpAddrConnect(addrlist, &fd)) == NULL) {
			error = errno;
			fd = -1;

			if (getenv("CLASS") != NULL) {
				printf("SocketOpenDevice() : INFO: Unable to contact printer, queuing on next "
							"printer in class...\n");
				/*
				 * Sleep 5 seconds to keep the job from requeuing too rapidly...
				 */

				sleep(5);

				return -1;
			}

			if (error == ECONNREFUSED || error == EHOSTDOWN ||
				error == EHOSTUNREACH) {
				if (contimeout && (time(NULL) - start_time) > contimeout) {
					printf("SocketOpenDevice() : ERROR: Printer not responding!\n");
					return -1;
				}

				recoverable = 1;

				printf("SocketOpenDevice() : WARNING: recoverable: Network host \'%s\' is busy; "
							"will retry in %d seconds...\n", hostname, delay);

				sleep(delay);

				if (delay < 30) {
					delay += 5;
				}
			} else {
				recoverable = 1;

				if (debug_mode) {
					printf("SocketOpenDevice() : DEBUG: Connection error: %s\n", strerror(errno));
				}
				printf("SocketOpenDevice() : ERROR: recoverable: Unable to connect to printer; "
							"will retry in 30 seconds...\n");
				sleep(30);
			}
		} else {
			break;
		}
	}

	if (recoverable) {
		fputs("SocketOpenDevice() : INFO: recovered:\n", stderr);
		sleep(5);
	}

	httpAddrFreeList(addrlist);

	return fd;
}

int getUSBDeviceURI (int fd, char *device_uri, int device_uriLen)
{
	char device_id[1024];
	char mfg[100];
	char mdl[100];
	char des[200];
	char *tmp;
	unsigned int device_idLen;
	unsigned int length;

	memset(device_id, 0, sizeof(device_id));
	memset(mfg, 0, sizeof(mfg));
	memset(mdl, 0, sizeof(mdl));
	
	if (device_uri == NULL || device_uriLen < 32) {
		printf("getUSBDeviceURI() : wrong argument.\n");

		return EXIT_FAILURE;
	}

	// get device id
	if (ioctl(fd, LPIOC_GET_DEVICE_ID(device_uriLen), device_id)) {
//		printf("getUSBDeviceURI() : failed to get device id.\n");

		return EXIT_FAILURE;
	}

	length = (((unsigned)device_id[0] & 255) << 8) +
		((unsigned)device_id[1] & 255);

/*
	if (length > device_idLen) {
		length = (((unsigned)device_id[1] & 255) << 8) +
			((unsigned)device_id[0] & 255);
	}

	if (length > device_idLen) {
		length = device_idLen;
	}
*/

	if (length > device_uriLen || length < 14) {
		length = (((unsigned)device_id[1] & 255) << 8) +
			((unsigned)device_id[0] & 255);
	}

	if (length > device_uriLen) {
		length = device_uriLen;
	}

	if (length < 14) {   
		printf("getUSBDeviceURI() : Icorrect device id.\n");

		return EXIT_FAILURE;
	} else {   
		length -= 2;

		memmove(device_id, device_id + 2, length);
		device_id[length] = '\0';
	}
	
	// get device option 
	// manufacture name, model name
	if (getDeviceOption(device_id, "MFG", mfg) == EXIT_FAILURE) {
		printf("getUSBDeviceURI() : failed to parse device id.\n");

		return EXIT_FAILURE;
	}

	if (getDeviceOption(device_id, "MDL", mdl) == EXIT_FAILURE) {
		if (getDeviceOption(device_id, "MODEL", mdl) == EXIT_FAILURE) {
			printf("getUSBDeviceURI() : failed to parse device id.\n");

			return EXIT_FAILURE;
		}
	}

	if (mfg == NULL) {
		if (getDeviceOption(device_id, "DES", des) == EXIT_FAILURE) {
			printf("getUSBDeviceURI() : failed to parse device id.\n");

			return EXIT_FAILURE;
		}

		if ((tmp = strchr(des, ' ')) != NULL)
			*tmp = '\0';

		strcpy(mfg, tmp);
	}

	if (!mdl) {
		memset(mdl, 0, sizeof(mdl));
	}


	if (!strncasecmp(mdl, mfg, strlen(mfg))) {
		tmp = mdl + strlen(mfg);

		while (isspace(*tmp & 255)) {
			tmp++;
		}

		memset(mdl, 0, sizeof(mdl));
		strcpy(mdl, tmp);

		tmp = mdl + strlen(mdl) - 1;

		while (isspace(*tmp & 255)) {
			*tmp = '\0';
			tmp--;
		}
	}

	tmp = mfg + strlen(mfg) - 1;

	while (isspace(*tmp & 255)) {
		*tmp = '\0';
		tmp--;
	}

	sprintf(device_uri, "usb://%s/%s", mfg, mdl);

	return EXIT_SUCCESS;
}

// 현재 device_id 형태, 표현방식이 바뀌면 이 함수도 수정되어야한다.
// device_id = MFG:BIXOLON ;CMD:BXL_01;MDL:SRP-350IIK;CLS:PRINTER;DES:BIXOLON_SRP-350IIK
int getDeviceOption (
		const char *device_id, 
		const char *option, 
		char *buffer)
{
	char delimiters[] = " \t";
	char *new_id;
	char *sp;
	char *value;

	if ((new_id = calloc(1, (strlen(device_id) + 1) * sizeof(char))) == NULL) {
		printf("getDeviceOption() : failed to memory allocation.\n");

		return EXIT_FAILURE;
	}

	strcpy(new_id, device_id);

cntParsing:

	if ((sp = strtok(new_id, ";")) == NULL) {
	//	printf("getDeviceOption() : failed to parse device id.\n");

		free(new_id);
		return EXIT_FAILURE;
	}

	sp = sp + strspn(sp, delimiters);

	if (strncmp(sp, option, strlen(option)) == 0) {
		if ((value = strchr(sp, ':')) == NULL) {
	//		printf("getDeviceOption() : failed to parse device id.\n");

			free(new_id);
			return EXIT_FAILURE;
		} else {
			value += 1;
			strcpy(buffer, value);

			free(new_id);
			return EXIT_SUCCESS;
		}
	} else {
		new_id = NULL;
		goto cntParsing;
	}
}

/****************************************************/
/******************  For Bluetooth  *****************/
/****************************************************/

int spp_connect (bdaddr_t *src, bdaddr_t *dst, uint8_t channel, 
				const char *cups_class, int *sk)
{
	struct sockaddr_rc addr;

	if ((*sk = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
		printf("Can't create bluetooth socket.\n");

		if (cups_class) {
			return CUPS_BACKEND_FAILED;
		} else {
			return CUPS_BACKEND_RETRY;
		}
	}

	addr.rc_family = AF_BLUETOOTH;
	bacpy(&addr.rc_bdaddr, src);
	addr.rc_channel = 0;

	if (bind(*sk, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		printf("Can't bind bluetooth socket.\n");
		close(*sk);
		*sk = -1;

		if (cups_class) {
			return CUPS_BACKEND_FAILED;
		} else {
			return CUPS_BACKEND_RETRY;
		}
	}

	addr.rc_family = AF_BLUETOOTH;
	bacpy(&addr.rc_bdaddr, dst);
	addr.rc_channel = channel;

	if (debug_mode) {
		printf("src.b[0] = %u, src.b[1] = %u, src.b[2] = %u, src.b[3] = %u, src.b[4] = %u, src.b[5] = %u\n",
				src->b[0], src->b[1], src->b[2], src->b[3], src->b[4], src->b[5]); 

		printf("dst.b[0] = %u, dst.b[1] = %u, dst.b[2] = %u, dst.b[3] = %u, dst.b[4] = %u, dst.b[5] = %u\n",
				dst->b[0], dst->b[1], dst->b[2], dst->b[3], dst->b[4], dst->b[5]); 

		printf("channel 	= %u\n", channel);
		printf("cups_class  = %s\n", cups_class);
		printf("sk 			= %d\n", *sk);
	}

reconnect:

	if (connect(*sk, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		printf("sppConnect() : Can't connect to device\n");
		close(*sk);
		*sk = -1;

		
		if (debug_mode) {
			printf("errno = %d\n", errno);
			if (errno == EBADF) {
				printf("connect error EBADF.\n");
			} else if (errno == EFAULT) { 
				printf("Connect error EFAULT.\n");
			} else if (errno == ENOTSOCK) { 
				printf("Connect error ENOTSOCK.\n");
			} else if (errno == EISCONN) {
				printf("Connect error EISCONN.\n");
			} else if (errno == ECONNREFUSED) {
				printf("Connect error CONNREFUSED.\n");
			} else if (errno == ETIMEDOUT) {
				printf("Connect error ETIMEOUT.\n");
			} else if (errno == ENETUNREACH) {
				printf("Connect error ENETUNREACH.\n");
			} else if (errno == EADDRINUSE) { 
				printf("Connect error EADDRINUSE.\n");
			} else if (errno == EINPROGRESS) { 
				printf("Connect error EINPROGRESS.\n");
			} else if (errno == EALREADY) {
				printf("Connect error EALREADY.\n");
			} else if (errno == EAGAIN) { 
				printf("Connect error EAGAIN.\n");
			} else if (errno == EAFNOSUPPORT) { 
				printf("Connect error EAFNOSUPPORT.\n");
			} else if (errno == EACCES) {
				printf("Connect error EACCES.\n");
			}
		}

		if (errno == EBUSY) {
			printf("Bluetooth device is busy. Reconnect after 10 secs.\n");

			sleep(10);

			goto reconnect;
		}

		if (cups_class) {
			return CUPS_BACKEND_FAILED;
		} else {
			return CUPS_BACKEND_RETRY;
		}
	}

	return CUPS_BACKEND_OK;
}

int sdp_search_spp(sdp_session_t *sdp, uint8_t *channel)
{
	sdp_list_t *srch, *attrs, *rsp;
	uuid_t svclass;
	uint16_t attr;
	int err;

	if (!sdp)
		return -1;

	// SERIAL_PORT_SVCLASS_ID(0x1101)를 data로 하는 uuid_t 객체 생성
	sdp_uuid16_create(&svclass, SERIAL_PORT_SVCLASS_ID);
	// 위에서 생성한 uuid_t를 data로 하는 sdp_list_t 객체 생성
	srch = sdp_list_append(NULL, &svclass);

	// srch에 data는 svclass(uuid_t)가 들어가 있음. next는 0
	// svclass는 type에 SDP_UUID16, value.uuid16에 SERIAL_PORT_SVCLASS_ID
	attr = SDP_ATTR_PROTO_DESC_LIST;
	attrs = sdp_list_append(NULL, &attr);

	// attrs에 data는 SDP_ATTR_PROTO_DESC_LIST, next는 0

	// SERIAL_PORT_SVCLASS_ID의 service session을 찾아서 
	// rsp로 받아온다.
	err = sdp_service_search_attr_req(
						sdp, 					 // session, 연결된 session
						srch, 					 // search, 찾을 service
						SDP_ATTR_REQ_INDIVIDUAL, // reqtype,  
						attrs, 					 // attrids
						&rsp);					 // rsp, 찾은 service

	if (err)
		return -1;

	for (; rsp; rsp = rsp->next) {
		sdp_record_t *rec = (sdp_record_t *) rsp->data;
		sdp_list_t *protos;

		if (!sdp_get_access_protos(rec, &protos)) {

			uint8_t ch = sdp_get_proto_port(protos, RFCOMM_UUID);

			if (ch > 0) {
				*channel = ch;

				return 0;
			}
		}
	}

	return -1;
}

/****************************************************/
/****************************************************/
