
/*
 *   Copyright (C) 2011-2012 Bixolon Co., Ltd.
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *        
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *             
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * BIXOLON CO LTD
 *
 * Thanks to people who are involved in this project.

 * Developed By dwlee  May-02-2012 Bixolon Co Ltd, South Korea
 * Mail: dwlee@bixolon.com
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include "bxlThermalCupsAPI.h"
#include "bxlMsq.h"
#include "bxlApiMsqKey.h"

#define MAXMSQLEN  4096
#define MSQENDMSG  "BxlMSQSendFinish"
#define BXLAPIKEY  "BIXOLON_API_MODE" 

static int readMsqID 	= -1;
static int writeMsqID 	= -1;

// XXX_AGENT_API_MSQKEY
static int readKey 	= AGENT_API_MSQKEY;
// XXX_API_FILTER_MSQKEY
static int writeKey	= API_FILTER_MSQKEY;

char fakeFile[100];
static int initFlag;
static char *SELECTED_PRINTER;

int modelNum;


/********************************************************/
/********************************************************/

static int createFakeFile (const char *printerName, char *fakeFile)
{
	FILE *fp;

	if (printerName == NULL) {
		printf("Printer name is NULL\n");

		return EXIT_FAILURE;
	}

	sprintf(fakeFile, ".BixolonAPI_%s", printerName);

	if ((fp = fopen(fakeFile, "w")) == NULL) {
		printf("failed to create file\n");

		return EXIT_FAILURE;
	}

	fprintf(fp, "Please delete printer from cups system and "
			    "reinstall the Bixolon printer CUPS driver.\n");

	if (fp != NULL) {
		fclose(fp);
	}

	return EXIT_SUCCESS;
}


static int bxlAPIInitialization () 
{
	cups_dest_t *dests;
	cups_dest_t *dest;

	int numDests;
	int i;

	if (SELECTED_PRINTER == NULL) {
		printf("printer name is NULL.\n");

		return EXIT_FAILURE;
	}

	memset(fakeFile, 0, sizeof(fakeFile));

	numDests = cupsGetDests(&dests);
	if (numDests < 1) {
		printf("bxlAPIInitialization() : "
				"Number of destinations is zero.\n");

		return EXIT_FAILURE;
	}

//	printf("SELECTED_PRINTER : %s\n", SELECTED_PRINTER);
	dest = cupsGetDest(SELECTED_PRINTER, NULL, numDests, dests);
	if (dest == NULL) {
		printf("bxlAPIInitialization() : cupsGetDest() faile.\n");
		printf("Please recheck selected printer name.\n");

		return EXIT_FAILURE;
	}

	for (i = 0; i < dest->num_options; i++) {
		if (strcmp(dest->options[i].name, "printer-make-and-model") == 0) {
			if (strcmp(dest->options[i].value, MAKENMODEL)) {
				printf("make and model : %s, %s\n", dest->options[i].value, MAKENMODEL);
				printf("bxlAPIInitialization() : "
						"This printer is not BIXOLON PRINTER.\n");

				return EXIT_FAILURE;
			}

			break;
		}
	}

	if ((readMsqID = bxlMsqOpen(readKey, MAXMSQLEN)) == EXIT_FAILURE) {
		printf("bxlAPIInitialization() : read bxlMsqOpen() fail.\n");

		return EXIT_FAILURE;
	}

	if ((writeMsqID = bxlMsqOpen(writeKey, MAXMSQLEN)) == EXIT_FAILURE) {
		printf("bxlAPIInitialization() : write bxlMsqOpen() fail.\n");

		return EXIT_FAILURE;
	}

	bxlMsqClean(writeMsqID);

	if (createFakeFile(SELECTED_PRINTER, fakeFile) == EXIT_FAILURE) {
		printf("printer name is incorrect.\n");

		return EXIT_FAILURE;
	}

	if (cupsPrintFile(dest->name, fakeFile, BXLAPIKEY,
					  dest->num_options, dest->options) == 0) {
		printf("bxlAPIInitialization() : "
				"failed to call Bixolon Cups Driver.\n");

		return EXIT_FAILURE;
	}

	/*
	if (bxlMsqSend(writeKey, SELECTED_PRINTER , strlen(SLECTED_PRINTER)) == EXIT_FAILURE) {
		printf("failed to send message queue.\n");

		return EXIT_FAILURE;
	}
	*/

	return EXIT_SUCCESS;
}


int startBxlAPIMode (const char *printerName)
{
	int len;

	len = strlen(printerName) + 1;

//	printf("len = %d, printerName = %s\n", len, printerName);
	if ((SELECTED_PRINTER = calloc(len, sizeof(char))) == NULL) {
		printf("failed memory allocation.\n");

		return EXIT_FAILURE;
	}

	memcpy(SELECTED_PRINTER, printerName, len);
//	printf("SLECTED_PRINTER ==== %s\n", SELECTED_PRINTER);

	return EXIT_SUCCESS;
}

int bxlCmdWriter (const char *buffer, int bufferLen)
{
	unsigned int len;
	char *bp;

	len = 0;
	bp 	= NULL;

	if (buffer == NULL || bufferLen < 0) {// ||
//		strlen(buffer) != bufferLen) {
		printf("Wrong buffer data.\n");

		return EXIT_FAILURE;
	}

//	printf("initFlag = %d\n", initFlag);
	if (initFlag == 0) {
		if (bxlAPIInitialization() == EXIT_FAILURE) {
			return EXIT_FAILURE;
		} else {
			initFlag = 1;
		}
	} else {
//		bxlMsqClean(writeMsqID);
	}


//	printf("bxlCmdWriter() buffer = %s, bufferLen = %d\n", buffer, bufferLen);

	if (bufferLen > MAXMSQLEN - sizeof(long)) {
		while (bufferLen > 0) {
			bp  = buffer + len;
			len = bufferLen > MAXMSQLEN - sizeof(long) ? 
								MAXMSQLEN - sizeof(long) : bufferLen;
			bufferLen -= len;

			if (bxlMsqSend(writeMsqID, writeKey, bp , len) == EXIT_FAILURE) {
				printf("failed to send message queue.\n");

				return EXIT_FAILURE;
			}
		}
	} else {
		if (bxlMsqSend(writeMsqID, writeKey, buffer, bufferLen) == EXIT_FAILURE) {
			printf("failed to send message queue.\n");

			return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}

int bxlReader (char *buffer, const int timeout)
{
	struct bxlMsg_t *msg;
	int len; 
	int count;

	msg 	= NULL;
	len 	= 0;
	count 	= 0; 

	if (buffer == NULL) {
		printf("incorrect parameter.\n");

		return EXIT_FAILURE;
	}

	while (count <= timeout * 100) { 

		if (timeout != 0) {
			count++;
		}

		msg = bxlMsqRecv(readMsqID, readKey, MAXMSQLEN, &len);

		if (msg == NULL) { 
			if (len == 0) { 
				usleep(10000);

				continue;
			} else {
				printf("printer status read fail.\n");

				return EXIT_FAILURE;
			}    
		} else {
			break;
		}    
	}    

//	printf("bxl API len = %d, strlen = %d\n", len, strlen(msg->mtext));
	if (msg != NULL) {
		memcpy(buffer, msg->mtext, len - sizeof(long));
		free(msg);

		return EXIT_SUCCESS;
	} else {
		printf("Empty msq.\n");

		return EXIT_FAILURE;
	} 
}

int finishBxlAPIMode ()
{
	char buff[50];
	int ret;
	
	initFlag = 0;

	memset(buff, 0, sizeof(buff));
	
	/* delete fake file */
	sprintf(buff, "rm -f .BixolonAPI_%s", fakeFile);
	system(buff);

	ret = bxlMsqSend(writeMsqID, writeKey, MSQENDMSG, strlen(MSQENDMSG));

	usleep(500000);

	if (SELECTED_PRINTER != NULL) {
		free(SELECTED_PRINTER);
	}

	return EXIT_SUCCESS;

}
